/*****************************************************************************
** STAIR VISION LIBRARY
** Copyright (c) 2007-2009, Stephen Gould
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are met:
**     * Redistributions of source code must retain the above copyright
**       notice, this list of conditions and the following disclaimer.
**     * Redistributions in binary form must reproduce the above copyright
**       notice, this list of conditions and the following disclaimer in the
**       documentation and/or other materials provided with the distribution.
**     * Neither the name of the Stanford University nor the
**       names of its contributors may be used to endorse or promote products
**       derived from this software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
** EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
** DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
** DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
** (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
** LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
** ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
** SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**
******************************************************************************
** FILENAME:    svlHistogram.cpp
** AUTHOR(S):   Olga Russakovsky <olga@cs.stanford.edu>
** DESCRIPTION:
**
*****************************************************************************/

#include "svlHistogram.h"

using namespace std;

// svlHistogram -----------------------------------------------------------------

svlHistogram::svlHistogram(unsigned numBins) :
  svlHistogramMemManager(numBins, true), _ops(numBins, 0, numBins)
{
  // do nothing
}

svlHistogram::svlHistogram(unsigned numBins, double maxValue, double minValue) :
  svlHistogramMemManager(numBins, true), _ops(numBins, minValue, maxValue)
{
  // do nothing
}

svlHistogram::svlHistogram(unsigned numBins, vector<double> *v, unsigned offset) :
  svlHistogramMemManager(numBins, false),  _ops(numBins, 0, numBins)
{
  if (v) set(v, offset);
}

svlHistogram::svlHistogram(unsigned numBins, double maxValue, double minValue, 
			   vector<double> *v, unsigned offset) :
  svlHistogramMemManager(numBins, false), _ops(numBins, minValue, maxValue)
{
  if (v) set(v, offset);
}

bool svlHistogram::insert(double binVal, double weight)
{
  SVL_ASSERT_MSG(_values, "storage for the feature vector must be provided");
  int bin = _ops.getBinNum(binVal);
  if (bin < 0) return false;
  elem(bin) += weight;
  return true;
}

bool svlHistogram::insertLinearSmooth(double binVal, double weight)
{
  SVL_ASSERT_MSG(_values, "storage for the feature vector must be provided");
  svlSoftBinAssignment sb;
  if (!_ops.getBinNumLinearSmooth(binVal, sb)) return false;

  elem(sb._bin[0]) += weight * sb._binWeight[0];
  elem(sb._bin[1]) += weight * sb._binWeight[1];
  return true;
}

// svl2dHistogram -----------------------------------------------------------------

svl2dHistogram::svl2dHistogram(unsigned numBins1, unsigned numBins2) :
  svlHistogramMemManager(numBins1*numBins2, true),
  _opsDim1(numBins1, 0, numBins1),
  _opsDim2(numBins2, 0, numBins2),
  _numBins2(numBins2)
{
  // do nothing
}

svl2dHistogram::svl2dHistogram(unsigned numBins1, double minValue1, double maxValue1,
			       unsigned numBins2, double minValue2, double maxValue2) :
  svlHistogramMemManager(numBins1*numBins2, true),
  _opsDim1(numBins1, minValue1, maxValue1),
  _opsDim2(numBins2, minValue2, maxValue2),
  _numBins2(numBins2)
{
  // do nothing
}

svl2dHistogram::svl2dHistogram(unsigned numBins1, unsigned numBins2, 
			       vector<double> *v, unsigned offset) :
  svlHistogramMemManager(numBins1*numBins2, false),
  _opsDim1(numBins1, 0, numBins1),
  _opsDim2(numBins2, 0, numBins2),
  _numBins2(numBins2)
{
  if (v) set(v, offset);
}

svl2dHistogram::svl2dHistogram(unsigned numBins1, double minValue1, double maxValue1,
			       unsigned numBins2, double minValue2, double maxValue2,
			       vector<double> *v, unsigned offset) :
  svlHistogramMemManager(numBins1*numBins2, true),
  _opsDim1(numBins1, minValue1, maxValue1),
  _opsDim2(numBins2, minValue2, maxValue2),
  _numBins2(numBins2)
{
  if (v) set(v, offset);
}

bool svl2dHistogram::insert(double binVal1, double binVal2, double weight)
{
  SVL_ASSERT_MSG(_values, "storage for the feature vector must be provided");
  int bin1 = _opsDim1.getBinNum(binVal1);
  if (bin1 < 0) return false;
  int bin2 = _opsDim2.getBinNum(binVal2);
  if (bin2 < 0) return false;

  elem2d(bin1, bin2) += weight;
  return true;
}

bool svl2dHistogram::insertLinearSmooth(double binVal1, double binVal2,
					double weight)
{
  SVL_ASSERT_MSG(_values, "storage for the feature vector must be provided");
  svlSoftBinAssignment sb1;
  if (!_opsDim1.getBinNumLinearSmooth(binVal1, sb1)) return false;

  svlSoftBinAssignment sb2;
  if (!_opsDim2.getBinNumLinearSmooth(binVal2, sb2)) return false;
  
  // insert values into all of the four possible bins
  for (unsigned i = 0; i < 2; i++) {
    for (unsigned j = 0; j < 2; j++) {
      double w = weight * sb1._binWeight[i] * sb2._binWeight[j];
      elem2d(sb1._bin[i], sb2._bin[j]) += w;
    }
  }

  return true; 
}

bool svl2dHistogram::insertLinearSmooth(svlSoftBinAssignment sb1, double binVal2,
					double weight)
{
  SVL_ASSERT_MSG(_values, "storage for the feature vector must be provided");

  svlSoftBinAssignment sb2;
  if (!_opsDim2.getBinNumLinearSmooth(binVal2, sb2)) return false;
  
  // insert values into all of the four possible bins
  for (unsigned i = 0; i < 2; i++) {
    for (unsigned j = 0; j < 2; j++) {
      double w = weight * sb1._binWeight[i] * sb2._binWeight[j];
      elem2d(sb1._bin[i], sb2._bin[j]) += w;
    }
  }

  return true;
}

