/*****************************************************************************
** STAIR VISION LIBRARY
** Copyright (c) 2007-2009, Stephen Gould
** All rights reserved.
**
** Redistribution and use in source and binary forms, with or without
** modification, are permitted provided that the following conditions are met:
**     * Redistributions of source code must retain the above copyright
**       notice, this list of conditions and the following disclaimer.
**     * Redistributions in binary form must reproduce the above copyright
**       notice, this list of conditions and the following disclaimer in the
**       documentation and/or other materials provided with the distribution.
**     * Neither the name of the Stanford University nor the
**       names of its contributors may be used to endorse or promote products
**       derived from this software without specific prior written permission.
**
** THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS ``AS IS'' AND ANY
** EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
** WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
** DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDERS BE LIABLE FOR ANY
** DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
** (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
** LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
** ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
** (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
** SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
**
******************************************************************************
** FILENAME:    svlPRcurve.h
** AUTHOR(S):   Olga Russakovsky <olga@cs.stanford.edu>
**              Ian Goodfellow <ia3n@cs.stanford.edu>
** DESCRIPTION:
**  Provides helper functions for analyzing a PR curve
**
*****************************************************************************/

#pragma once

#include <cassert>
#include <iostream>
#include <vector>
#include <map>

#include "cv.h"
#include "cxcore.h"
#include "ml.h"

#include "svlBase.h"

using namespace std;

typedef enum _svlPRpointSort_t {
  NONE, BY_RECALL, BY_PRECISION, BY_FSCORE
} svlPRpointSort_t;

// helper class: svlPRpoint

class svlPRpoint {
 private:
  float _threshold;
  float _recall;
  float _pr;
  float _Fscore;
  
 public:
  svlPRpointSort_t _sort_flag;

 public:
  svlPRpoint() : _sort_flag(NONE) {}
  svlPRpoint(float i, float j, float k) :
    _threshold(i), _recall(j), _pr(k), _sort_flag(NONE) {}
  
    inline bool contains_nan() const {
      return ((_pr != _pr) || (_recall != _recall));
    }

    bool operator<(const svlPRpoint &a) const;
    void print(ofstream &ofs) const;
    float getFscore() const;
    void saveFscore();

    // computes the contribution to the pr curve from the
    // previous point on the curve to the current
    // (assumes the the points have already been sorted)
    float getPRcontribution(const svlPRpoint *prev = NULL) const;

    // determines if this point is a duplicate on the pr curve given
    // what the _next_ point is; the points should already be sorted,
    // so they will be in order of, say, increasing recall, and within
    // the same recall value they will be in order of increasing
    // precision
    bool isDuplicate(const svlPRpoint &next) const;
};

// [[TODO: verify that all of this works with arbitrary scores, not just probs]]

typedef vector<svlPRpoint> svlPRcurve;

void writePRcurve(const char *filestem, const svlPRcurve &curve);
bool readPRcurve(const char *filename, svlPRcurve &curve);
float getHighestFscore(const svlPRcurve &curve);
 // use only if have a preference for PR vs. Recall sorting;
// by default will average sorting by precision and by recall
float getPRarea(const svlPRcurve &curve, svlPRpointSort_t s = NONE);

svlPRcurve makePRcurve(vector<double> & posScores, vector<double> & negScores);
